;*******************************************************************************
;* These are functions needed to do minimal video interface in protected mode. *
;* They include an interrupt handler which implements only BIOS video function *
;* 0EH, the TTY character write, and two functions which use that function to  *
;* display a string, and to display the contents of the eax register in hex.   *
;*******************************************************************************

;For use in V86.ASM, etc.

;(C) 1993 American Eagle Publications, Inc., All rights reserved!


;This procedure displays the null terminated string at DS:SI on the console.
DISPLAY_MSG:    lodsb
                or      al,al
                jz      DM_EXIT
                mov     bh,0
                mov     ah,0EH
                push    esi
                CALL    VIDEO_SERVICE
                pop     esi
                jmp     DISPLAY_MSG
DM_EXIT:        ret

;Display the number in ax in hex format
DISPLAY_NUMBER:
                mov     cl,32
DNLP:           push    ecx
                push    eax
                call    DISPLAY_DIGIT
                pop     eax
                pop     ecx
                sub     cl,4
                or      cl,cl
                jnz     DNLP
                ret

DISPLAY_DIGIT:
                sub     cl,4
                shr     eax,cl
                and     eax,0FH
                mov     ebx,eax
                add     ebx,OFFSET HEX_DIGIT
                mov     al,[ebx]
                mov     ah,0EH
                mov     ebx,0
                CALL    VIDEO_SERVICE
                ret

HEX_DIGIT       DB      '0123456789ABCDEF'


;*******************************************************************************
; Protected mode BIOS starts here.
;*******************************************************************************

                ASSUME  DS:BIOS_DATA

;Make sure VIDEO_SEG is properly defined or these won't work!
ILLEGAL_ROW     EQU     25              ;First number of illegal video row

;Position the cursor on the active page to the correct location.
;ACTIVE_CURSOR:
;                xor     ebx,ebx
;                mov     bl,[ACTIVE_PAGE]
;                sal     ebx,1                   ;index for getting cursor position
;                mov     ax,[OFFSET C_POS + bx]  ;ah = row, al = column
;                mov     cl,al
;                xor     ch,ch                   ;cx = column
;                mov     bx,[SCR_WIDTH]          ;get screen width in bx
;                mov     al,ah
;                xor     ah,ah
;                mul     bx                      ;ax = screen width * rows
;                add     cx,ax                   ;cx = cursor address in linear memory
;                mov     dx,[PORT_6845]          ;get controller port address
;                mov     al,0EH                  ;set high byte of cursor position
;                out     dx,al
;                inc     edx
;                mov     al,ch
;                out     dx,al
;                dec     edx
;                mov     al,0FH                  ;and then low byte
;                out     dx,al
;                inc     edx
;                mov     al,cl
;                out     dx,al
;                ret


;BIOS Video Service Number 10
;This is identical to service 9, except that it leaves the color attribute
;which is currently in video memory untouched.
WRITE_CHAR:
                mov     ah,bl
                push    eax                             ;save character and attribute
                xor     edi,edi
                mov     bl,bh
                xor     bh,bh
                mov     ax,[REGEN_LENGTH]               ;calculate address of page
                mul     bx
                xor     edi,edi
                mov     di,ax                           ;correct page offset here
                sal     ebx,1                           ;index to C_POS array for this page
                mov     bx,[OFFSET C_POS + bx]          ;bx = cursor position in array
                mov     al,bh
                xor     ah,ah
                mov     dx,[SCR_WIDTH]
                mul     dx                              ;ax = rows * width
                xor     bh,bh
                add     eax,ebx                         ;ax = columns + rows * width
                sal     eax,1                           ;multiply by 2 for word array
                add     di,ax                           ;di = offset in memory of character to write
                pop     eax                             ;restore ax
WC1:            stosb                                   ;and go write one byte
                ret

;BIOS Video Service Number 14
;This routine puts a character in memory and advances the cursor one position.
;If necessary, it advances to the next line, scrolls up, etc. The character
;is passed in AL. The Bell, Backspace, Line Feed and Carriage Return characters
;get special processing. Page is passed in BH.
WRITE_CHAR_TTY:
                cmp     al,0AH                  ;is it line feed?
                je      SHORT WCT_LF
                cmp     al,0DH                  ;is it carriage return?
                jne     SHORT WCT_1
                jmp     SHORT WCT_CR
WCT_1:          push    ebx                     ;save page number
                call    WRITE_CHAR
                pop     ebx
                mov     bl,bh
                xor     bh,bh                   ;bx = page number
                sal     bx,1
                mov     ax,[OFFSET C_POS + bx]  ;get cursor position for this page
                inc     al
                cmp     al,BYTE PTR [SCR_WIDTH] ;is column = screen width?
                jne     SHORT WCT_2             ;no - just increment it by one
                xor     al,al                   ;yes - zero the column
                inc     ah                      ;and increment the row count
WCT_2:          cmp     ah,ILLEGAL_ROW          ;have we exceeded the maximum row number?
                jne     SHORT WCT_3             ;no - update cursor and exit
                dec     ah                      ;go back one row
                push    eax
                push    ebx
                call    SCROLL_PAGE_SCREEN      ;and scroll the page up one line
                pop     ebx
                pop     eax
WCT_3:          mov     [OFFSET C_POS + bx],ax  ;save the new cursor position
;                call    ACTIVE_CURSOR           ;and update active page's cursor
                retn

WCT_LF:                                 ;handle line feed
                mov     bl,bh
                xor     bh,bh
                sal     bx,1
                mov     ax,[OFFSET C_POS + bx]  ;get cursor position for this page
                inc     ah
                jmp     WCT_2                   ;go scroll if needed, and update cursor

WCT_CR:                                 ;handle carriage return
                mov     bl,bh
                xor     bh,bh
                sal     bx,1
                mov     ax,[OFFSET C_POS + bx]  ;get cursor position for this page
                xor     al,al                   ;clear column number
                mov     [OFFSET C_POS + bx],ax  ;store the cursor position
;                call    ACTIVE_CURSOR           ;and reposition the cursor
                ret


;This scrolls the whole screen up one line. The page to scroll is passed in BH.
SCROLL_PAGE_SCREEN:
                xor     eax,eax
                mov     bl,bh
                xor     bh,bh                   ;bx = page number
                mov     ax,[REGEN_LENGTH]
                mul     bx                      ;ax = offset to page
                mov     edi,eax                 ;destination index for move
                add     ax,[SCR_WIDTH]          ;index forward one line
                add     ax,[SCR_WIDTH]
                mov     esi,eax                 ;source index for move
                mov     ax,[SCR_WIDTH]
                mov     ecx,24                  ;24 lines to move
                mul     cx                      ; * characters per line
                mov     ecx,eax                 ;number of words to move
                push    ds
                push    es
                pop     ds
                rep     movsw                   ;move display up
                pop     ds
                xor     ecx,ecx
                mov     cx,[SCR_WIDTH]          ;and clear last line
                mov     ax,256*7+32             ;using spaces with attribute 7
                rep     stosw
                retn

;Substitute BIOS video interrupt for protected mode. Handles only function 14!
VIDEO_SERVICE:
                push    ds
                push    es
;                push    ebx
;                push    ecx
;                push    edx
;                push    esi
;                push    edi
                push    eax
                mov     ax,BIOS_SEL
                mov     ds,ax
                mov     ax,VIDEO_SEL
                mov     es,ax
                pop     eax
VIDEO_E:        cmp     ah,0EH                  ;Service 14, write character as tty device
                jne     SHORT VIDEO_EXIT
                call    WRITE_CHAR_TTY
VIDEO_EXIT:
;                pop     edi
;                pop     esi
;                pop     edx
;                pop     ecx
;                pop     ebx
                pop     es
                pop     ds
                ret
